"""
Module that implements the observer pattern.
"""

##################################
#    Generic Observable class    #
##################################

class Events():
    """
    Class that represents the events that can be observed.
    """
    
    SEND_BEAT_EVENT                 = "SEND_BEAT_EVENT"
    """
    Used to notify that beats must be sent.
    """
    END_CONNECTION_EVENT            = "END_CONNECTION_EVENT"
    """
    Used to notify that a connection has been closed. (arg: NodeConnection)
    """
    AGREGATION_FINISHED_EVENT       = "AGREGATION_FINISHED_EVENT"
    """
    Used to notify that the agregation was done. (arg: model or None)
    """
    CONN_TO_EVENT                   = "CONN_TO_EVENT"
    """
    Used to notify when a node must connect to other. (arg: (host,port))
    """
    START_LEARNING_EVENT            = "START_LEARNING_EVENT"
    """
    Used to notify when the learning process starts. (arg: (rounds,epochs))
    """
    STOP_LEARNING_EVENT             = "STOP_LEARNING_EVENT"
    """
    Used to notify when the learning process stops.
    """
    PARAMS_RECEIVED_EVENT           = "PARAMS_RECEIVED_EVENT"
    """
    Used to notify when the parameters are received. (arg: params (encoded))
    """
    METRICS_RECEIVED_EVENT          = "METRICS_RECEIVED_EVENT"
    """
    Used to notify when the metrics are received. (arg: (node, round, loss, metric))
    """
    TRAIN_SET_VOTE_RECEIVED_EVENT   = "TRAIN_SET_VOTE_RECEIVED_EVENT"
    """
    Used to notify when a vote is received. (arg: (node,votes))
    """
    NODE_CONNECTED_EVENT            = "NODE_CONNECTED_EVENT"
    """
    Used to notify when a node is connected. (arg: (n, force))
    """
    PROCESSED_MESSAGES_EVENT        = "PROCESSED_MESSAGES_EVENT"
    """
    Used to notify when a node process messages. (arg: (node, messages))
    """
    GOSSIP_BROADCAST_EVENT          = "GOSSIP_BROADCAST_EVENT"
    """
    Used to notify when a node must send gossiped messages. (arg: (msg,nodes))
    """
    BEAT_RECEIVED_EVENT             = "BEAT_RECEIVED_EVENT"
    """
    Used to notify when a node receive a beat. (arg: node)
    """
    
##################################
#    Generic Observable class    #
##################################

class Observable():
    """
    Class that implements the **Observable** at the observer pattern.
    """
    def __init__(self):
        self.__observers = []

    def add_observer(self, observer):
        """
        Adds an observer to the list of observers.

        Args:
            observer: The observer to add.
        """
        self.__observers.append(observer)
    
    def get_observers(self):
        """
        Returns the list of observers.
        
        Returns:
            The list of observers.
        """
        return self.__observers

    def notify(self, event, obj):
        """
        Notifies to all the observers an event.

        Args:
            event: The event to notify.
            obj: The object to pass to the observer. For each event, the object is different (check it at the ``Event`` class).
        """
        [o.update(event, obj) for o in self.__observers]

##################################
#    Generic Observer class    #
##################################

class Observer():
    """
    Class for the **Observer** at the observer pattern.
    
    Args:
        event: The event that is notified.
        obj: The object that is passed by the observable.
    """
    def update(self, event, obj): pass


import logging
import threading
import time
from p2pfl.settings import Settings
from p2pfl.utils.observer import Events, Observable

#####################
#    Heartbeater    #
#####################

class Heartbeater(threading.Thread, Observable):
    """
    Thread based heartbeater that sends a beat message to all the neighbors of a node every `HEARTBEAT_PERIOD` seconds.

    It also maintains a list of active neighbors, which is created by receiving different heartbear messages. 
    Neighbors from which a heartbeat is not received in ``Settings.NODE_TIMEOUT`` will be eliminated

    Communicates with node via observer pattern.

    Args:
        nodo_padre (Node): Node that use the heartbeater.
    """
    def __init__(self, node_name):
        Observable.__init__(self)
        threading.Thread.__init__(self, name = "heartbeater-" + node_name)
        self.__node_name = node_name
        self.__terminate_flag = threading.Event()

        # List of neighbors
        self.__nodes = {}

    def run(self):
        """
        Send a beat every HEARTBEAT_PERIOD seconds to all the neighbors of the node. 
        Also, it will clear from the neighbors list the nodes that haven't sent a heartbeat in NODE_TIMEOUT seconds. 
        It happend ``HEARTBEATER_REFRESH_NEIGHBORS_BY_PERIOD`` per HEARTBEAT_PERIOD 
        """
        while not self.__terminate_flag.is_set():
            # We do not check if the message was sent
            #   - If the model is sending, a beat is not necessary
            #   - If the connection its down timeouts will destroy connections
            self.notify(Events.SEND_BEAT_EVENT, None)
            # Wait and refresh node list
            for _ in range(Settings.HEARTBEATER_REFRESH_NEIGHBORS_BY_PERIOD):
                self.clear_nodes()
                time.sleep(Settings.HEARTBEAT_PERIOD/Settings.HEARTBEATER_REFRESH_NEIGHBORS_BY_PERIOD)


    def clear_nodes(self):
        """
        Clear the list of neighbors.
        """
        for n in [node for node,t in list(self.__nodes.items()) if time.time() - t > Settings.NODE_TIMEOUT]:
            logging.debug("({}) Removed {} from the network ".format(self.__node_name, n))
            self.__nodes.pop(n)

    def add_node(self, node):
        """
        Add a node to the list of neighbors.

        Args:
            node (Node): Node to add to the list of neighbors.
        """
        if node != self.__node_name:
            self.__nodes[node] = time.time()

    def get_nodes(self):
        """
        Get the list of actual neighbors.

        Returns:
            list: List of neighbors.
        """
        node_list = list(self.__nodes.keys())
        if self.__node_name not in node_list:
            node_list.append(self.__node_name)
        return node_list

    def stop(self):
        """
        Stop the heartbeater.
        """
        self.__terminate_flag.set()



class DecodingParamsError(Exception):  
    """
    An exception raised when decoding parameters fails.
    """
    pass

class ModelNotMatchingError(Exception):  
    """
    An exception raised when parameters not match with the model.
    """
    pass


from Crypto import Random
from Crypto.PublicKey import RSA
from Crypto.Cipher import PKCS1_OAEP
import base64
from Crypto import Random
from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes

class Encrypter:
    """
    Class with methods to encrypt and decrypt messages.
    """

    def encrypt(self, message):
        pass

    def decrypt(self, message):      
        pass

    def get_key(self):
        pass
    
###############################
#    Asymmetric Encryption    #
###############################

class RSACipher(Encrypter):
    """
    Class with methods to encrypt and decrypt messages using RSA asymetric encryption.
    """

    def __init__(self):
        random_generator = Random.new().read
        self.__private_key = RSA.generate(1024, random_generator)
        self.__public_key = self.__private_key.publickey()
        self.__pair_public_key = None

    def encrypt(self, message):
        """
        Encrypts a message using RSA. Message is encrypted using the public key of the pair (the other node key).
        
        Args:
            message: (bytes) The message to encrypt.

        Returns:
            message: (bytes) The encrypted message.
        """
        cipher = PKCS1_OAEP.new(self.__pair_public_key)
        return base64.b64encode(cipher.encrypt(message))

    def decrypt(self, message):
        """
        Decrypts a message using RSA. Message is decripted using the private key. 

        Args:
            message: (bytes) The message to decrypt.

        Returns:
            message: (bytes) The decrypted message.
        """
        cipher = PKCS1_OAEP.new(self.__private_key)
        return cipher.decrypt(base64.b64decode(message))

    def load_pair_public_key(self, key):
        """
        Loads the public key of the pair (other node).

        Args:
            key: The key to use to decrypt the message encoded at base64.
        """
        self.__pair_public_key = RSA.importKey(base64.b64decode(key))

    def get_key(self):
        """
        Get the serialized RSA public key.
        
        Returns:
            key: The serialized key encoded at base64.
        """
        return base64.b64encode(self.__public_key.exportKey("DER"))

##############################
#    Symmetric Encryption    #
##############################

class AESCipher(Encrypter):
    """
    Class with methods to encrypt and decrypt messages using AES symetric encryption.
    """
    
    def __init__(self, key=None): 
        self.bs = AES.block_size
        self.key = key
        if key is None:
            self.key = get_random_bytes(16) # 256 bits
        self.cipher = AES.new(self.key, AES.MODE_ECB)

    def encrypt(self, message):
        """
        Encrypts a message using AES. Message is encrypted using the shared key.
        Keep in mind that AES uses a block cipher, so the message sould be filled with padding.

        Args:
            message: (str) The message to encrypt.

        Returns:
            message: (str) The encrypted message. 
        """
        return self.cipher.encrypt(message)

    def decrypt(self, message):
        """
        Decrypts a message using AES. Message is decripted using the shared key. 
        Keep in mind that AES uses a block cipher, so the message can be a filled message with padding.

        Args:
            message: (bytes) The message to decrypt.

        Returns:
            message: (bytes) The decrypted message.
        """
        return self.cipher.decrypt(message)
    
    def add_padding(self, msg):
        """
        Add padding to a encoded UTF-8 text. Adds " " charactets (1 byte) to fill the rest of the block.
        Careful: for this case the filling content don't affect to the meaning of messages.

        Args:
            msg: (bytes) The encoded text.
        
        Returns:
            msg: (bytes) The encoded text with padding.
        """
        # Calculate the number of bytes needed to fill the block
        bytes_left = self.bs - len(msg) % self.bs
        if bytes_left == self.bs:
            bytes_left = 0

        # Add padding
        pading_char = " ".encode("utf-8")
        for i in range(bytes_left):
            msg += pading_char

        return msg 

    def get_key(self):
        """
        Get the shared RSA key.
        
        Returns:
            key: The shared key.
        """
        return self.key

    def get_block_size():
        """
        Returns:
            block_size: (int) The length of the block in bytes.
        """
        return AES.block_size

    def key_len():
        """ 
        Returns:
            key_len: (int) The length of the key in bytes.
        """
        return 256


import os
from pytorch_lightning.loggers.base import LightningLoggerBase, rank_zero_experiment
from pytorch_lightning.utilities.distributed import rank_zero_only
from torch.utils.tensorboard import SummaryWriter
import torch

class FederatedTensorboardLogger(LightningLoggerBase):
    """
    Logger for PyTorch Lightning in federated learning. 
    It is made to save the information of the different trainings (in the different rounds) in the same graph.

    End of round determined with `finalize_round`

    Args:
        dir (str): Directory where the logs will be saved.
        name (str): Name of the node.
        version (int): Version of the experiment.
    """

    def __init__(self, dir, name = None, version=0, **kwargs):
        super().__init__()
        self._name = "unknown_node"
        self._version = version
        if name is not None:
            self._name = name

        # Create log directory
        dir = os.path.join(dir, self._name)
        # If exist the experiment, increment the version
        while os.path.exists(os.path.join(dir, "experiment_" + str(version))):
            version += 1
        # Create the writer
        self.writer = SummaryWriter(os.path.join(dir, "experiment_" + str(version)))

        # FL information
        self.round = 0
        self.local_step = 0
        self.global_step = 0
        
        self.writer.add_scalar("fl_round", self.round, self.global_step)

        
    @property
    def name(self):
        """
        """
        return self._name

    @property
    def version(self):
        """
        """
        return self._version

    @rank_zero_only
    def log_hyperparams(self, params):
        """
        """
        # params is an argparse.Namespace
        # your code to record hyperparameters goes here
        pass

    @rank_zero_only
    def log_metrics(self, metrics, step):
        """
        """

        # FL round information
        self.local_step = step
        __step = self.global_step + self.local_step

        # Log Round
        self.writer.add_scalar("fl_round", self.round, __step)
       
        for k, v in metrics.items():
            if isinstance(v, torch.Tensor):
                v = v.item()

            if isinstance(v, dict):
                self.writer.add_scalars(k, v, __step)
            else:
                try:
                    self.writer.add_scalar(k, v, __step)
                # todo: specify the possible exception
                except Exception as ex:
                    m = f"\n you tried to log {v} which is currently not supported. Try a dict or a scalar/tensor."
                    raise ValueError(m) from ex

    def log_scalar(self, key, value, round=None, name=None):
        """
        """
        if round is None:
            round = self.round

        if name is None:
            self.writer.add_scalar(key, value, round)
        else:
            self.writer.add_scalars(key, {name: value}, round)


    @rank_zero_only
    def save(self):
        """
        """
        # Optional. Any code necessary to save logger data goes here
        pass

    def finalize(self, status):
        """
        """
        pass

    def finalize_round(self):
        # Finish Round
        self.global_step = self.global_step + self.local_step
        self.local_step = 0
        self.round = self.round + 1

    def close(self):
        try:
            self.writer.close()
        except:
            pass


"""
Module to define constants for the p2pfl system.
"""

import logging
from p2pfl.encrypter import AESCipher

###################
# Global Settings #
###################

class Settings():
    """
    Class to define global settings for the p2pfl system.
    """

    # TCP Block Size
    BLOCK_SIZE = 2048
    """
    Max ammount of bytes to read from the buffer at a time. If ```BLOCK_SIZE`` is not divisible by the block size used for symetric encryption it will be rounded to the next closest value.

    Try to strike a balance between hyper-segmentation and excessively large block size.
    """
    rest = BLOCK_SIZE % AESCipher.get_block_size()
    if rest != 0:
        new_value = BLOCK_SIZE + AESCipher.get_block_size() - rest
        logging.info("Changing buffer size to %d. %d is incompatible with the AES block size.", BLOCK_SIZE, new_value)
        BLOCK_SIZE = new_value

    # TIMEOUT'S
    NODE_TIMEOUT = 20
    """
    Timeout (seconds) for a node to be considered dead.
    """
    VOTE_TIMEOUT = 60
    """
    Timeout (seconds) for a node to wait for a vote.
    """
    AGREGATION_TIMEOUT = 60
    """
    Timeout (seconds) for a node to wait for other models. Timeout stars when the first model is added.
    """

    # HEARTBEATER
    HEARTBEAT_PERIOD = 4
    """
    Period (seconds) for the node to send a heartbeat.
    """
    HEARTBEATER_REFRESH_NEIGHBORS_BY_PERIOD = 4
    """
    Times by period to refresh the neighbors list.
    """
    WAIT_HEARTBEATS_CONVERGENCE = 10
    """
    Time (seconds) to wait for the heartbeats to converge.
    """
    
    # TRAIN SET 
    TRAIN_SET_SIZE = 10
    """
    Size of the training set.
    """
    TRAIN_SET_CONNECT_TIMEOUT = 5
    """
    Timeout (seconds) to wait for a node to connect to the training set.
    """

    # GOSSIP Messages
    AMOUNT_LAST_MESSAGES_SAVED = 100 
    """
    Amount of last messages saved. Used to avoid gossiping cycles.
    """
    GOSSIP_MESSAGES_FREC = 100
    """
    Frequency of gossiping messages (times by second).
    """
    GOSSIP_MESSAGES_PER_ROUND = 100
    """
    Amount of gossip messages per round.
    """

    # GOSSIP Models
    GOSSIP_EXIT_ON_X_EQUAL_ROUNDS = 20
    """
    Amount of equal rounds to exit gossiping. Careful, a low value can cause an early stop of gossiping.
    """
    GOSSIP_MODELS_FREC = 1
    """
    Frequency of gossiping models (times by second).
    """
    GOSSIP_MODELS_PER_ROUND = 2
    """
    Amount of gossip models per round.
    """
    FRAGMENTS_DELAY = 0.0
    """
    Delay (seconds) to wait before sending a fragment. This is a very important value, if the node is too slow and the buffer isn't big enough, the node will send fragments too fast and the other node will not receive them.
    """


import math
import random
import threading
import logging
import time
from p2pfl.base_node import BaseNode
from p2pfl.command import *
from p2pfl.communication_protocol import CommunicationProtocol
from p2pfl.settings import Settings
from p2pfl.learning.agregators.fedavg import FedAvg
from p2pfl.learning.exceptions import DecodingParamsError, ModelNotMatchingError
from p2pfl.learning.pytorch.lightninglearner import LightningLearner
from p2pfl.utils.observer import Events, Observer

class Node(BaseNode):
    """
    Class based on a base node that allows **p2p Federated Learning**. 
    
    Metrics will be saved under a folder with the name of the node.
            
    Args:
        model: Model to be learned. Carefull, model should be compatible with data and the learner.
        data: Dataset to be used in the learning process. Carefull, model should be compatible with data and the learner.
        host (str): Host where the node will be listening.
        port (int): Port where the node will be listening.
        learner: Learner to be used in the learning process. Default: LightningLearner.
        agregator (Agregator): Agregator to be used in the learning process. Default: FedAvg.
        simulation (bool): If False, node will share metrics and communication will be encrypted. Default: True.

    Attributes:
        round (int): Round of the learning process.
        totalrounds (int): Total number of rounds of the learning process.
        learner (Learner): Learner to be used in the learning process.
        agregator (Agregator): Agregator to be used in the learning process.
    """

    #####################
    #     Node Init     #
    #####################

    def __init__(self, model, data, host="127.0.0.1", port=None, learner=LightningLearner, agregator=FedAvg, simulation=True):
        # Super init
        BaseNode.__init__(self,host,port, simulation)
        Observer.__init__(self)

        # Learning
        self.round = None 
        self.totalrounds = None 
        self.__model_initialized = False
        self.__initial_neighbors = []
        self.__start_thread_lock = threading.Lock()

        # Learner
        self.learner = learner(model, data, log_name=self.get_name()) 

        # Agregator
        self.agregator = agregator( node_name = self.get_name() )
        self.agregator.add_observer(self)

        # Train Set Votes
        self.__train_set = []
        self.__train_set_votes = {}
        self.__train_set_votes_lock = threading.Lock()

        # Locks
        self.__wait_votes_ready_lock = threading.Lock()
        self.__finish_agregation_lock = threading.Lock()
        self.__finish_agregation_lock.acquire()
        self.__wait_init_model_lock = threading.Lock()
        self.__wait_init_model_lock.acquire()

    #########################
    #    Node Management    #
    #########################

    def connect_to(self, h, p, full=False, force=False):
        """"
        Connects a node to other. If learning is running connections are not allowed (it should be forced).
        Carefull, if connection is forced with a new node, it will produce timeouts in the network.
    
        Args:
            h (str): The host of the node.
            p (int): The port of the node.
            full (bool): If True, the node will be connected to the entire network.
            force (bool): If True, the the node will be connected even though it should not be.

        Returns:
            node: The node that has been connected to.
        """
        # Check if learning is running
        if self.round is not None and not force:
            logging.info("({}) Cant connect to other nodes when learning is running.".format(self.get_name()))
            return None

        # Connect
        return super().connect_to(h, p, full, force)

    def stop(self): 
        """
        Stop the node and the learning if it is running.
        """
        if self.round is not None:
            self.__stop_learning()
        self.learner.close()
        super().stop()

    ##########################
    #    Learning Setters    #
    ##########################

    def set_data(self, data):
        """
        Set the data to be used in the learning process (learner).

        Args:
            data: Dataset to be used in the learning process.
        """
        self.learner.set_data(data)

    def set_model(self, model):
        """"
        Set the model to use. 
        Carefully, model, not only weights.

        Args:
            model: Model to be learned.
        """
        self.learner.set_model(model)

    ###############################################
    #         Network Learning Management         #
    ###############################################

    def set_start_learning(self, rounds=1, epochs=1): 
        """
        Start the learning process in the entire network.

        Args:
            rounds: Number of rounds of the learning process.
            epochs: Number of epochs of the learning process.
        """
        if self._terminate_flag.is_set():
            logging.info("({}) Node must be running to start learning".format(self.get_name()))
            return
        if self.round is None:
            # Start Learning
            logging.info("({}) Broadcasting start learning...".format(self.get_name()))
            self.broadcast(CommunicationProtocol.build_start_learning_msg(rounds,epochs))
            # Initialize model
            self.broadcast(CommunicationProtocol.build_model_initialized_msg())
            self.__wait_init_model_lock.release()
            self.__model_initialized = True # esto seguramente sobre, con locks es suficiente
            # Learning Thread
            self.__start_learning_thread(rounds,epochs)
        else:
            logging.info("({}) Learning already started".format(self.get_name()))

    def set_stop_learning(self):
        """
        Stop the learning process in the entire network.
        """
        if self.round is not None:
            self.broadcast(CommunicationProtocol.build_stop_learning_msg())
            self.__stop_learning()
        else:
            logging.info("({}) Learning already stopped".format(self.get_name()))


    ##################################
    #         Local Learning         #
    ##################################

    def __start_learning_thread(self,rounds,epochs):
        learning_thread = threading.Thread(target=self.__start_learning,args=(rounds,epochs))
        learning_thread.name = "learning_thread-" + self.get_name()
        learning_thread.daemon = True
        learning_thread.start()

    def __start_learning(self,rounds,epochs):
        """
        Start the learning process in the local node.
        
        Args:
            rounds: Number of rounds of the learning process.
            epochs: Number of epochs of the learning process.
        """
        self.__start_thread_lock.acquire() # Used to avoid create duplicated training threads
        if self.round is None:
            self.round = 0
            self.totalrounds = rounds
            self.learner.init()
            self.__start_thread_lock.release()

            begin = time.time()

            # Wait and gossip model inicialization            
            self.__gossip_model_difusion(initialization=True)

            # Wait to guarantee new connection heartbeats convergence and fix neighbors
            wait_time = Settings.WAIT_HEARTBEATS_CONVERGENCE - (time.time() - begin)
            if wait_time > 0:
                time.sleep(wait_time)
            self.__initial_neighbors = self.get_neighbors() # used to restore the original list of neighbors after the learning round

            # Train
            self.learner.set_epochs(epochs)
            self.__train_step()

    def __stop_learning(self): 
        """
        Stop the learning process in the local node. Interrupts learning process if its running.
        """
        logging.info("({}) Stopping learning".format(self.get_name()))
        # Rounds
        self.round = None
        self.totalrounds = None
        # Leraner
        self.learner.interrupt_fit()
        # Agregator
        self.agregator.check_and_run_agregation(force=True)  
        self.agregator.set_nodes_to_agregate([])
        self.agregator.clear()
        # Try to free wait locks
        try:
            self.__wait_votes_ready_lock.release()
        except:
            pass

    ####################################
    #         Model Agregation         #
    ####################################

    def add_model(self,m): 
        """
        Add a model. If the model isn't inicializated, the recieved model is used for it. Otherwise, the model is agregated using the **agregator**.

        Args:
            m: Encoded model. Contains model and their contributors
        """
        # Check if Learning is running
        if self.round is not None:
            try:
                if self.__model_initialized:
                    # Add model to agregator
                    decoded_model, contributors, weight = self.learner.decode_parameters(m)
                    if self.learner.check_parameters(decoded_model):
                        models_added = self.agregator.add_model(decoded_model,contributors,weight)
                        if models_added is not None:
                            # CAREFULL RARE BUG at MACBOOCK: When CPU is high, only new nodes will be sent.
                            self.broadcast(CommunicationProtocol.build_models_agregated_msg(models_added))
                    else:
                        raise ModelNotMatchingError("Not matching models")
                else:
                    # Initialize model
                    model, _, _ = self.learner.decode_parameters(m)
                    self.learner.set_parameters(model)
                    self.__model_initialized = True
                    logging.info("({}) Initialicing Model Weights".format(self.get_name()))
                    self.__wait_init_model_lock.release()    
                    self.broadcast(CommunicationProtocol.build_model_initialized_msg())
            
            except DecodingParamsError as e:
                logging.error("({}) Error decoding parameters".format(self.get_name()))
                self.stop()

            except ModelNotMatchingError as e:
                logging.error("({}) Models not matching.".format(self.get_name()))
                self.stop()
                    
            except Exception as e:
                self.stop()
                raise(e)
        else: 
            logging.error("({}) Tried to add a model while learning is not running".format(self.get_name()))
    
    #######################
    #    Trainig Steps    #
    #######################

    def __train_step(self):

        # Set train set
        if self.round is not None:
            self.__train_set = self.__vote_train_set()
            self.__validate_train_set()

        # Determine if node is in the train set
        is_train_set = self.get_name() in self.__train_set
        if is_train_set:

            # Full connect train set
            if self.round is not None:
                self.__connect_and_set_agregator()

            # Evaluate and send metrics
            if self.round is not None:
                self.__evaluate()

            # Train
            if self.round is not None:
                self.__train()
            
            # Agregate Model
            if self.round is not None:
                self.agregator.add_model(self.learner.get_parameters(),[self.get_name()],self.learner.get_num_samples()[0])
                self.broadcast(CommunicationProtocol.build_models_agregated_msg([self.get_name()])) # Notify agregation
                self.__gossip_model_agregation()
        else: 

            # Set Models To Agregate 
            self.agregator.set_waiting_agregated_model()

        # Gossip agregated model (also syncrhonizes nodes)
        if self.round is not None:
            self.__gossip_model_difusion()

        # Finish round
        if self.round is not None:
            self.__on_round_finished()

    ################
    #    Voting    #
    ################

    def __vote_train_set(self):

        # Vote
        candidates = self.get_network_nodes() # al least himself
        logging.debug("({}) {} candidates to train set".format(self.get_name(),len(candidates)))
        if self.get_name() not in candidates:
            candidates.append(self.get_name())

        # Send vote
        samples = min(Settings.TRAIN_SET_SIZE,len(candidates))
        nodes_voted = random.sample(candidates, samples)
        weights = [math.floor(random.randint(0,1000)/(i+1)) for i in range(samples)]
        votes = list(zip(nodes_voted,weights))

        # Adding votes
        self.__train_set_votes_lock.acquire()
        self.__train_set_votes[self.get_name()] = dict(votes)
        self.__train_set_votes_lock.release()

        # Send and wait for votes
        logging.info("({}) Sending train set vote.".format(self.get_name()))
        logging.debug("({}) Self Vote: {}".format(self.get_name(),votes))
        self.broadcast(CommunicationProtocol.build_vote_train_set_msg(self.get_name(),votes))
        logging.debug("({}) Waiting other node votes.".format(self.get_name()))

        # Get time
        count = 0
        begin = time.time()                

        while True:
            # If the trainning has been interrupted, stop waiting
            if self.round is None:
                logging.info("({}) Stopping on_round_finished process.".format(self.get_name()))
                return []

            # Update time counters (timeout)
            count = count + (time.time() - begin)
            begin = time.time()                
            timeout = count > Settings.VOTE_TIMEOUT

            # Clear non candidate votes
            self.__train_set_votes_lock.acquire()
            nc_votes = {k:v for k,v in self.__train_set_votes.items() if k in candidates}
            self.__train_set_votes_lock.release()
                
            # Determine if all votes are received
            votes_ready = set(candidates) == set(nc_votes.keys())
            if votes_ready or timeout:

                if timeout and not votes_ready:
                    logging.info("({}) Timeout for vote agregation. Missing votes from {}".format(self.get_name(), set(candidates) - set(nc_votes.keys())))

                results = {}
                for node_vote in list(nc_votes.values()):
                    for i in range(len(node_vote)):
                        k = list(node_vote.keys())[i]
                        v = list(node_vote.values())[i]
                        if k in results:
                            results[k] += v
                        else:
                            results[k] = v

                # Order by votes and get TOP X
                results = sorted(results.items(), key=lambda x: x[0], reverse=True) # to equal solve of draw
                results = sorted(results, key=lambda x: x[1], reverse=True)
                top = min(len(results), Settings.TRAIN_SET_SIZE)
                results = results[0:top]
                results = {k: v for k, v in results}
                votes = list(results.keys())

                # Clear votes
                self.__train_set_votes = {}
                logging.info("({}) Computed {} votes.".format(self.get_name(),len(nc_votes)))
                return votes
                
            # Wait for votes or refresh every 2 seconds
            self.__wait_votes_ready_lock.acquire(timeout=2) 
        
                                
    def __validate_train_set(self):
        # Verify if node set is valid (can happend that a node was down when the votes were being processed)
        for tsn in self.__train_set:
            if tsn not in self.get_network_nodes():
                if tsn != self.get_name():
                    self.__train_set.remove(tsn)
                
        logging.info("{} Train set of {} nodes: {}".format(self.get_name(),len(self.__train_set),self.__train_set))

    ##########################
    #    Connect Trainset    #
    ##########################

    def __connect_and_set_agregator(self):
        # Set Models To Agregate 
        self.agregator.set_nodes_to_agregate(self.__train_set)

        # Connect Train Set Nodes
        for node in self.__train_set :
            if node != self.get_name():
                h,p = node.split(":")
                if p.isdigit():
                    nc = self.get_neighbor(h,int(p))
                    # If the node is not connected, connect it (to avoid duplicated connections only a node connects to the other)
                    if nc is None and self.get_name() > node: 
                        self.connect_to(h,int(p),force=True)
                else:
                    logging.info("({}) Node {} has a valid port".format(self.get_name(),node.split(":")))

        # Wait connections
        count = 0
        begin = time.time()
        while True:
            count = count + (time.time() - begin)
            if count > Settings.TRAIN_SET_CONNECT_TIMEOUT:
                logging.info("({}) Timeout for train set connections.".format(self.get_name()))
                break
            if len(self.__train_set) == len([nc for nc in self.get_neighbors() if nc.get_name() in self.__train_set])+1:
                break
            time.sleep(0.1) 


    ############################
    #    Train and Evaluate    #
    ############################

    def __train(self):
        logging.info("({}) Training...".format(self.get_name()))
        self.learner.fit()

    def __evaluate(self):
        logging.info("({}) Evaluating...".format(self.get_name()))
        results = self.learner.evaluate()
        if results is not None:
            logging.info("({}) Evaluated. Losss: {}, Metric: {}. (Check tensorboard for more info)".format(self.get_name(),results[0],results[1]))
            # Send metrics
            if not self.simulation:
                logging.info("({}) Broadcasting metrics.".format(self.get_name(),len(self.get_neighbors())))
                encoded_msgs = CommunicationProtocol.build_metrics_msg(self.get_name(),self.round,results[0],results[1])
                self.broadcast(encoded_msgs)

    ######################
    #    Round finish    #
    ######################

    def __on_round_finished(self):
        # Remove trainset connections
        for nc in self.get_neighbors():
            if nc not in self.__initial_neighbors:
                self.rm_neighbor(nc)
        # Set Next Round
        self.agregator.clear()
        self.learner.finalize_round() # revisar x si esto pueiera quedar mejor
        self.round = self.round + 1
        # Clear node agregation
        for nc in self.get_neighbors():
            nc.clear_models_agregated()

        # Next Step or Finish
        logging.info("({}) Round {} of {} finished.".format(self.get_name(),self.round,self.totalrounds))
        if self.round < self.totalrounds:
            self.__train_step()  
        else:
            # At end, all nodes compute metrics
            self.__evaluate()
            # Finish
            self.round = None
            self.totalrounds = None
            self.__model_initialized = False
            logging.info("({}) Training finished!!.".format(self.get_name(),self.round,self.totalrounds))

    #########################
    #    Model Gossiping    #
    #########################

    def __gossip_model_agregation(self):
        # Anonymous functions
        candidate_condition = lambda nc: nc.get_name() in self.__train_set and len(nc.get_models_agregated())<len(self.__train_set)
        status_function = lambda nc: ( nc.get_name(),len(nc.get_models_agregated()) )
        model_function = lambda nc: self.agregator.get_partial_agregation(nc.get_models_agregated())

        # Gossip
        self.__gossip_model(candidate_condition,status_function,model_function)
         
    def __gossip_model_difusion(self,initialization=False):
        # Wait a model (init or agregated)
        if initialization:
            logging.info("({}) Waiting initialization.".format(self.get_name()))
            self.__wait_init_model_lock.acquire()
            logging.info("({}) Gossiping model initialization.".format(self.get_name(), len(self.get_neighbors())))
            candidate_condition = lambda nc: not nc.get_model_initialized()
        else:
            logging.info("({}) Waiting aregation.".format(self.get_name()))
            self.__finish_agregation_lock.acquire()
            logging.info("({}) Gossiping agregated model.".format(self.get_name(), len(self.get_neighbors())))
            candidate_condition = lambda nc: nc.get_model_ready_status()<self.round

        # Anonymous functions
        status_function = lambda nc: nc.get_name()
        model_function = lambda _: (self.learner.get_parameters(),None, None) # At diffusion, contributors are not relevant
        
        # Gossip
        self.__gossip_model(candidate_condition,status_function,model_function)       

    def __gossip_model(self, candidate_condition, status_function, model_function):
        # Initialize list with status of nodes in the last X iterations
        last_x_status = [] 
        j = 0

        while True:
            # Get time to calculate frequency
            begin = time.time()

            # If the trainning has been interrupted, stop waiting
            if self.round is None:
                logging.info("({}) Stopping model gossip process.".format(self.get_name()))
                return

            # Get nodes wich need models
            nei = [nc for nc in self.get_neighbors() if candidate_condition(nc)]

            # Determine end of gossip
            if nei == []:
                logging.info("({}) Gossip finished.".format(self.get_name()))
                return

            # Save state of neightbors. If nodes are not responding gossip will stop
            if len(last_x_status) != Settings.GOSSIP_EXIT_ON_X_EQUAL_ROUNDS:
                last_x_status.append([status_function(nc) for nc in nei])
            else:
                last_x_status[j] = str([status_function(nc) for nc in nei])
                j = (j+1)%Settings.GOSSIP_EXIT_ON_X_EQUAL_ROUNDS

                # Check if las messages are the same
                for i in range(len(last_x_status)-1):
                    if last_x_status[i] != last_x_status[i+1]:
                        break
                    logging.info("({}) Gossiping exited for {} equal reounds.".format(self.get_name(), Settings.GOSSIP_EXIT_ON_X_EQUAL_ROUNDS))
                    return

            # Select a random subset of neightbors
            samples = min(Settings.GOSSIP_MODELS_PER_ROUND,len(nei))
            nei = random.sample(nei, samples)

            # Generate and Send Model Partial Agregations (model, node_contributors)
            for nc in nei:
                model,contributors,weights = model_function(nc)

                # Send Partial Agregation
                if model is not None:
                    logging.info("({}) Gossiping model to {}.".format(self.get_name(), nc.get_name()))
                    encoded_model = self.learner.encode_parameters(params=model, contributors=contributors, weight=weights)
                    encoded_msgs = CommunicationProtocol.build_params_msg(encoded_model)
                    # Send Fragments
                    for msg in encoded_msgs:
                        nc.send(msg)
                
            # Wait to guarantee the frequency of gossipping
            time_diff = time.time() - begin
            time_sleep = 1/Settings.GOSSIP_MODELS_FREC-time_diff
            if time_sleep > 0:
                time.sleep(time_sleep)


    ###########################
    #     Observer Events     #
    ###########################

    def update(self,event,obj):
        """
        Observer update method. Used to handle events that can occur in the different components and connections of the node.
        
        Args:
            event (Events): Event that has occurred.
            obj: Object that has been updated. 
        """
        if event == Events.NODE_CONNECTED_EVENT:
            n, force = obj
            if self.round is not None and not force:
                logging.info("({}) Cant connect to other nodes when learning is running. (however, other nodes can be connected to the node.)".format(self.get_name()))
                n.stop()
                return
                
        elif event == Events.AGREGATION_FINISHED_EVENT:
            # Set parameters and communate it to the training process
            if obj is not None:
                self.learner.set_parameters(obj)
                # Share that agregation is done
                self.broadcast(CommunicationProtocol.build_models_ready_msg(self.round))
            else:
                logging.error("({}) Agregation finished with no parameters".format(self.get_name()))
                self.stop()
            try:
                self.__finish_agregation_lock.release()
            except:
                pass

        elif event == Events.START_LEARNING_EVENT:
            self.__start_learning_thread(obj[0],obj[1])

        elif event == Events.STOP_LEARNING_EVENT:
            self.__stop_learning()
    
        elif event == Events.PARAMS_RECEIVED_EVENT:
            self.add_model(obj)
        
        elif event == Events.METRICS_RECEIVED_EVENT:
            name, round, loss, metric = obj
            self.learner.log_validation_metrics(loss,metric,round=round,name=name)

        elif event == Events.TRAIN_SET_VOTE_RECEIVED_EVENT:
            node,votes = obj
            self.__train_set_votes_lock.acquire()
            self.__train_set_votes[node] = votes
            self.__train_set_votes_lock.release()
            # Communicate to the training process that a vote has been received
            try:
                self.__wait_votes_ready_lock.release()
            except:
                pass

        # Execute BaseNode update
        super().update(event,obj)


"""
Module that implements commands of the command pattern.
"""

#################
#    Command    #
#################

class Command:
    """
    Class that represents a command.

    Args:
        node_connection: The node connection that is going to execute the command.
    """
    def __init__(self, node_connection):
        self.node_connection = node_connection

    def execute(self,*args): pass

########################
#       Commands       #
########################

class Beat_cmd(Command):
    """
    Command that should be executed as a response to a **beat** message.
    """
    def execute(self, node1):
        self.node_connection.notify_heartbeat(node1)

class Stop_cmd(Command):
    """
    Command that is should be executed as a response to a **stop** message.
    """
    def execute(self):
        self.node_connection.stop(local=True)

class Conn_to_cmd(Command):
    """
    Command that should be executed as a response to a **conn_to** message.
    """
    def execute(self,h,p):
        self.node_connection.notify_conn_to(h,p)

class Start_learning_cmd(Command):
    """
    Command that should be executed as a response to a **start_learning** message.
    """
    def execute(self, rounds, epochs):
        self.node_connection.notify_start_learning(rounds, epochs)

class Stop_learning_cmd(Command):
    """
    Command that should be executed as a response to a **stop_learning** message.
    """
    def execute(self):
        self.node_connection.notify_stop_learning()

class Params_cmd(Command):
    """
    Command that should be executed as a response to a **params** message.
    """
    def execute(self,msg,done):
        self.node_connection.add_param_segment(msg)
        if done:
            params = self.node_connection.get_params()
            self.node_connection.clear_buffer()
            self.node_connection.notify_params(params)

class Models_Ready_cmd(Command):
    """
    Command that should be executed as a response to a **ready** message.
    """
    def execute(self,round):
        self.node_connection.set_model_ready_status(round)

class Metrics_cmd(Command):
    """
    Command that should be executed as a response to a **metrics** message.
    """
    def execute(self, node, round, loss, metric):
        self.node_connection.notify_metrics(node,round,loss,metric)


class Vote_train_set_cmd(Command):
    """
    Command that should be executed as a response to a **vote** message.
    """
    def execute(self, node, votes):
        self.node_connection.notify_train_set_votes(node, votes)

class Models_agregated_cmd(Command):
    """
    Command that should be executed as a response to a **models_agregated** message.
    """
    def execute(self,node_list):
        self.node_connection.add_models_agregated(node_list)

class Model_initialized_cmd(Command):
    """
    Command that should be executed as a response to a **model_initialized** message.
    """
    def execute(self):
        self.node_connection.set_model_initialized(True)


import logging
import torch
from p2pfl.learning.agregators.agregator import Agregator
   
class FedAvg(Agregator):
    """
    Federated Averaging (FedAvg) [McMahan et al., 2016]
    Paper: https://arxiv.org/abs/1602.05629
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def agregate(self,models): 
        """
        Ponderated average of the models.

        Args:
            models: Dictionary with the models (node: model,num_samples).
        """

        # Check if there are models to agregate
        if len(models)==0:
            logging.error("({}) Trying to agregate models when there is no models".format(self.node_name))
            return None

        models = list(models.values())
        
        # Total Samples
        total_samples = sum([y for _,y in models])

        # Create a Zero Model
        accum = (models[-1][0]).copy()
        for layer in accum:
            accum[layer] = torch.zeros_like(accum[layer])

        # Add weighteds models
        for m,w in models:
            for layer in m:
                accum[layer] = accum[layer] + m[layer]*w

        # Normalize Accum
        for layer in accum:
            accum[layer] = accum[layer]/total_samples

        return accum


from setuptools import find_packages, setup
from pathlib import Path

HERE = Path(__file__).parent

PACKAGE_NAME = 'p2pfl' 
VERSION = '0.1.6' 
AUTHOR = 'Pedro Guijas' 
AUTHOR_EMAIL = 'pguijas@gmail.com' 
URL = 'https://pguijas.github.io/federated_learning_p2p/' 
DESCRIPTION = 'p2p Federated Learning framework' 
LONG_DESCRIPTION = (HERE / "README.md").read_text(encoding='utf-8')
LONG_DESC_TYPE = "text/markdown"
LICENSE = 'MIT' 

INSTALL_REQUIRES = [
    "torch==1.11.0",
    "tensorboard",
    "pytorch-lightning",
    "torchvision",
    "pycryptodome"
] # dependencias


setup(
    name=PACKAGE_NAME, 
    version=VERSION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type=LONG_DESC_TYPE,
    license=LICENSE,
    install_requires=INSTALL_REQUIRES,
    package_dir={"p2pfl": "p2pfl"},
    packages=find_packages(
        where=".",
        include=[
            "*",
        ],
        exclude=[
            "test",
        ],
    ),
    include_package_data=True,
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    test_suite='test'
)





import torch
from torch import nn
from torch.nn import functional as F
import pytorch_lightning as pl
from torchmetrics import Accuracy


###############################
#    Multilayer Perceptron    #
###############################

class MLP(pl.LightningModule):
    """
    Multilayer Perceptron (MLP) to solve MNIST with PyTorch Lightning.
    """

    def __init__(self, metric = Accuracy, out_channels=10, lr_rate=0.001, seed=None): # low lr to avoid overfitting
        
        # Set seed for reproducibility iniciialization
        if seed is not None:
            torch.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)
            
        super().__init__()
        self.lr_rate = lr_rate
        self.metric = metric()

        self.l1 = torch.nn.Linear(28 * 28, 128)
        self.l2 = torch.nn.Linear(128, 256)
        self.l3 = torch.nn.Linear(256, out_channels)

    def forward(self, x):
        """
        """
        batch_size, channels, width, height = x.size()
        
        # (b, 1, 28, 28) -> (b, 1*28*28)
        x = x.view(batch_size, -1)
        x = self.l1(x)
        x = torch.relu(x)
        x = self.l2(x)
        x = torch.relu(x)
        x = self.l3(x)
        x = torch.log_softmax(x, dim=1)
        return x

    def configure_optimizers(self):
        """
        """
        return torch.optim.Adam(self.parameters(), lr=self.lr_rate)

    def training_step(self, batch, batch_id):
        """
        """
        x, y = batch
        loss = F.cross_entropy(self(x), y)
        self.log("train_loss", loss, prog_bar=True)
        return loss 

    def validation_step(self, batch, batch_idx):
        """
        """
        x, y = batch
        logits = self(x)
        loss = F.cross_entropy(self(x), y)
        out = torch.argmax(logits, dim=1)
        metric = self.metric(out, y) 
        self.log("val_loss", loss, prog_bar=True)
        self.log("val_metric", metric, prog_bar=True)    
        return loss

    def test_step(self, batch, batch_idx):
        """
        """
        x, y = batch
        logits = self(x)
        loss = F.cross_entropy(self(x), y)
        out = torch.argmax(logits, dim=1)
        metric = self.metric(out, y) 
        self.log("test_loss", loss, prog_bar=True)
        self.log("test_metric", metric, prog_bar=True)
        return loss
        


from collections import OrderedDict
import pickle
import torch
from pytorch_lightning import Trainer
from p2pfl.learning.learner import NodeLearner
from p2pfl.learning.exceptions import DecodingParamsError, ModelNotMatchingError
from p2pfl.learning.pytorch.logger import FederatedTensorboardLogger
import logging

###########################
#    LightningLearner     #
###########################

class LightningLearner(NodeLearner):
    """
    Learner with PyTorch Lightning.

    Atributes:
        model: Model to train.
        data: Data to train the model.
        log_name: Name of the log.
        epochs: Number of epochs to train.
        logger: Logger.
    """
    def __init__(self, model, data, log_name=None):
        self.model = model            
        self.data = data
        self.log_name = log_name
        self.logger = None
        self.__trainer = None
        self.epochs = 1
        # To avoid GPU/TPU printings
        logging.getLogger("pytorch_lightning").setLevel(logging.WARNING)

    def set_model(self, model):
        self.model = model

    def set_data(self, data):
        self.data = data
        
    def encode_parameters(self, params=None, contributors=None, weight=None):
        if params is None:
            params = self.model.state_dict()
        array = [val.cpu().numpy() for _, val in params.items()]
        return pickle.dumps((array,contributors,weight))

    def decode_parameters(self, data):
        try:
            params, contributors, weight = pickle.loads(data)
            params_dict = zip(self.model.state_dict().keys(), params)
            return (OrderedDict({k: torch.tensor(v) for k, v in params_dict}), contributors, weight)
        except:
            raise DecodingParamsError("Error decoding parameters")

    def check_parameters(self, params):
        # Check ordered dict keys
        if set(params.keys()) != set(self.model.state_dict().keys()):
            return False
        # Check tensor shapes
        for key, value in params.items():
            if value.shape != self.model.state_dict()[key].shape:
                return False
        return True

    def set_parameters(self, params):
        try:
            self.model.load_state_dict(params)
        except:
            raise ModelNotMatchingError("Not matching models")

    def get_parameters(self):
        return self.model.state_dict()

    def set_epochs(self, epochs):
        self.epochs = epochs

    def fit(self):
        try:
            if self.epochs > 0:
                self.__trainer = Trainer(max_epochs=self.epochs, accelerator="auto", logger=self.logger, enable_checkpointing=False, enable_model_summary=False) 
                self.__trainer.fit(self.model, self.data)
                self.__trainer = None
        except Exception as e:
            logging.error("Something went wrong with pytorch lightning. {}".format(e))

    def interrupt_fit(self):
        if self.__trainer is not None:
            self.__trainer.should_stop = True
            self.__trainer = None
            
    def evaluate(self):
        try:
            if self.epochs > 0:
                self.__trainer = Trainer(max_epochs=self.epochs, accelerator="auto", logger=None, log_every_n_steps=0, enable_checkpointing=False)
                results = self.__trainer.test(self.model, self.data, verbose=False)
                loss = results[0]["test_loss"]
                metric = results[0]["test_metric"]
                self.__trainer = None
                self.log_validation_metrics(loss, metric)
                return loss,metric
            else:
                return None
        except Exception as e:
            logging.error("Something went wrong with pytorch lightning. {}".format(e))
            return None

    def log_validation_metrics(self, loss, metric, round=None, name=None):
        self.logger.log_scalar("test_loss", loss, round,name=name)
        self.logger.log_scalar("test_metric", metric, round,name=name)

    def get_num_samples(self):
        return (len(self.data.train_dataloader().dataset), len(self.data.test_dataloader().dataset))

    def init(self):
        self.close()
        self.logger = FederatedTensorboardLogger("training_logs", name=self.log_name)
   
    def close(self):
        if self.logger is not None:
            self.logger.close()

    def finalize_round(self):
        self.logger.finalize_round()